#include <ntstatus.h>

#define WIN32_LEAN_AND_MEAN
#define WIN32_NO_STATUS
#include <Windows.h>
#include <winternl.h>

#include <cstdio>
#include <cstdlib>
#include <cwchar>

#include "Pe.hpp"

#pragma comment(lib, "ntdll.lib")



template <typename FnDeleter>
class Deleter
{
private:
    FnDeleter m_deleter;

public:
    explicit Deleter(FnDeleter deleter) noexcept
        : m_deleter(deleter)
    {
    }

    ~Deleter() noexcept
    {
        m_deleter();
    }
};



NTSTATUS mapLib(HANDLE hProcess, const wchar_t* const path) noexcept
{
    const HANDLE hFile = CreateFileW(path, FILE_READ_DATA, 0, nullptr, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, nullptr);
    if (hFile == INVALID_HANDLE_VALUE)
    {
        return STATUS_FILE_NOT_AVAILABLE;
    }

    const Deleter fileCloser([hFile]()
    {
        CloseHandle(hFile);
    });

    const DWORD libSize = GetFileSize(hFile, nullptr);
    if (!libSize)
    {
        return STATUS_UNSUCCESSFUL;
    }

    void* const libFile = VirtualAlloc(nullptr, libSize, MEM_RESERVE | MEM_COMMIT, PAGE_EXECUTE_READWRITE);
    if (!libFile)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    const Deleter libFileDeleter([libFile]()
    {
        VirtualFree(libFile, 0, MEM_RELEASE);
    });

    DWORD readBytes = 0;
    const bool readStatus = ReadFile(hFile, libFile, libSize, &readBytes, nullptr);
    if (!readStatus)
    {
        return STATUS_UNSUCCESSFUL;
    }

    const auto peFile = Pe::Pe64::fromFile(libFile);

    const auto originalImageBase = peFile.imageBase();
    const auto imageSize = peFile.imageSize();

    // Allocate memory in the target process:
    void* const remoteImageBase = VirtualAllocEx(hProcess, nullptr, imageSize, MEM_RESERVE | MEM_COMMIT, PAGE_EXECUTE_READWRITE);
    if (!remoteImageBase)
    {
        return STATUS_UNSUCCESSFUL;
    }

    /* Don't free the remote memory */

    // Allocate the local buffer to perform relocation and imports resolving:
    void* const localImage = VirtualAlloc(nullptr, imageSize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
    if (!localImage)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    const Deleter localImageDeleter([localImage]()
    {
        VirtualFree(localImage, 0, MEM_RELEASE);
    });

    // Copy headers:
    memcpy(localImage, libFile, peFile.headers().opt()->SizeOfHeaders);

    // Copy sections:
    for (const auto& sec : peFile.sections())
    {
        memcpy(
            reinterpret_cast<void*>(reinterpret_cast<size_t>(localImage) + sec.VirtualAddress),
            reinterpret_cast<void*>(reinterpret_cast<size_t>(libFile) + sec.PointerToRawData),
            sec.SizeOfRawData
        );
    }

    // Resolve imports:
    auto peModule = Pe::Pe64::fromModule(localImage);
    for (auto& imp : peModule.imports())
    {
        const char* const libName = imp.libName();
        const HMODULE hModule = GetModuleHandleA(libName);
        if (!hModule)
        {
            return STATUS_PROCEDURE_NOT_FOUND;
        }

        for (auto& fn : imp)
        {
            const char* const procName = &fn.name()->Name[0];
            const auto procAddr = GetProcAddress(hModule, procName);
            if (!procAddr)
            {
                return STATUS_PROCEDURE_NOT_FOUND;
            }

            auto* const entry = const_cast<Pe::Types<Pe::Arch::x64>::ImportAddressTableEntry*>(fn.importAddressTableEntry());
            entry->raw = reinterpret_cast<unsigned long long>(procAddr);
        }
    }

    // Perform relocations:
    const auto loadDelta = reinterpret_cast<size_t>(remoteImageBase) - static_cast<size_t>(originalImageBase);
    for (const auto& relocPage : peModule.relocs())
    {
        for (const auto& reloc : relocPage)
        {
            void* const addr = const_cast<void*>(reloc.addr());
            const auto* const rel = reloc.reloc();
            const auto type = rel->type();

            switch (type)
            {
            case Pe::RelocType::absolute:
            {
                // Do nothing.
                break;
            }
            case Pe::RelocType::high:
            {
                *static_cast<unsigned short*>(addr) += HIWORD(loadDelta);
                break;
            }
            case Pe::RelocType::low:
            {
                *static_cast<unsigned short*>(addr) += LOWORD(loadDelta);
                break;
            }
            case Pe::RelocType::highlow:
            {
                *static_cast<unsigned int*>(addr) += static_cast<unsigned int>(loadDelta);
                break;
            }
            case Pe::RelocType::highadj:
            {
                // Do nothing.
                break;
            }
            case Pe::RelocType::dir64:
            {
                *static_cast<unsigned long long*>(addr) += loadDelta;
                break;
            }
            }
        }
    }

    // Write to the remote process:
    SIZE_T written = 0;
    const bool remoteWriteStatus = !!WriteProcessMemory(hProcess, remoteImageBase, localImage, imageSize, &written);
    if (!remoteWriteStatus)
    {
        return STATUS_UNSUCCESSFUL;
    }

    const auto entryPoint = reinterpret_cast<DWORD(*__stdcall)(HMODULE, DWORD, LPCVOID)>(
        peModule.entryPoint() - reinterpret_cast<size_t>(localImage) + reinterpret_cast<size_t>(remoteImageBase)
        );

    void* const remoteShell = VirtualAllocEx(hProcess, nullptr, 4096, MEM_RESERVE | MEM_COMMIT, PAGE_EXECUTE_READWRITE);
    if (!remoteShell)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    #pragma pack(push, 1)
    class Shell
    {
    private:
        const unsigned char pushRbp{ 0x55 };                                        // push rbp
        const unsigned char movRbpRsp[3]{ 0x48, 0x89, 0xE5 };                       // mov rbp, rsp
        const unsigned char movRaxRcx[3]{ 0x48, 0x89, 0xC8 };                       // mov rax, rcx
        const unsigned char movRcxOpcode[2]{ 0x48, 0xB9 };                          // movabs rcx, <uint64_t>
        unsigned long long base;
        const unsigned char movRdx1[7]{ 0x48, 0xC7, 0xC2, 0x01, 0x00, 0x00, 0x00 }; // mov rdx, DLL_PROCESS_ATTACH
        const unsigned char movR8_1[7]{ 0x49, 0xC7, 0xC0, 0x01, 0x00, 0x00, 0x00 }; // mov r8, 1
        const unsigned char subRsp20h[4]{ 0x48, 0x83, 0xEC, 0x20 };                 // sub rsp, 0x20
        const unsigned char callRax[2]{ 0xFF, 0xD0 };                               // call rax
        const unsigned char addRsp20h[4]{ 0x48, 0x83, 0xC4, 0x20 };                 // add rsp, 0x20
        const unsigned char ret[1]{ 0xC3 };                                         // ret

    public:
        explicit Shell(const void* moduleBase) noexcept
            : base(reinterpret_cast<size_t>(moduleBase))
        {
        }
    };
    #pragma pack(pop)

    const Shell localShell(remoteImageBase);
    written = 0;
    const bool shellWriteStatus = !!WriteProcessMemory(hProcess, remoteShell, &localShell, sizeof(localShell), &written);
    if (!shellWriteStatus)
    {
        return STATUS_UNSUCCESSFUL;
    }

    DWORD tid = 0;
    const HANDLE hThread = CreateRemoteThread(hProcess, nullptr, 0, reinterpret_cast<LPTHREAD_START_ROUTINE>(remoteShell), entryPoint, 0, &tid);
    if (!hThread)
    {
        return STATUS_UNSUCCESSFUL;
    }

    printf("TID: %u\n", tid);

    WaitForSingleObject(hThread, INFINITE);
    CloseHandle(hThread);

    return STATUS_SUCCESS;
}



int wmain(int argc, wchar_t* argv[])
{
    LoadLibraryW(L"user32.dll");

    if (argc != 3)
    {
        printf("Usage:\n");
        printf("%ws \"N:\\Folder\\Lib.dll\" 1234\n", argv[0]);
        printf("  Where '1234' is the PID of the target process.\n");
        return 0;
    }

    const wchar_t* const libPath = argv[1];
    const unsigned int pid = static_cast<unsigned int>(std::wcstoull(argv[2], nullptr, 10));

    const HANDLE hProcess = OpenProcess(PROCESS_VM_OPERATION | PROCESS_VM_WRITE | PROCESS_CREATE_THREAD, false, pid);
    if (!hProcess)
    {
        return GetLastError();
    }

    const NTSTATUS status = mapLib(hProcess, libPath);
    printf("0x%X\n", status);

    return (status == ERROR_SUCCESS) ? 0 : -1;
}
